package ch.psi.daq.queryrest.query;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map.Entry;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import javax.annotation.Resource;

import org.apache.commons.lang3.tuple.Pair;
import org.apache.commons.lang3.tuple.Triple;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.context.ApplicationContext;

import ch.psi.daq.domain.DataEvent;
import ch.psi.daq.domain.backend.BackendAccess;
import ch.psi.daq.domain.config.DomainConfig;
import ch.psi.daq.domain.json.ChannelName;
import ch.psi.daq.domain.json.channels.info.ChannelInfos;
import ch.psi.daq.domain.query.ChannelNameRequest;
import ch.psi.daq.domain.query.DAQQueries;
import ch.psi.daq.domain.query.DAQQueryElement;
import ch.psi.daq.domain.query.backend.BackendQuery;
import ch.psi.daq.domain.query.backend.BackendQueryImpl;
import ch.psi.daq.domain.query.backend.analyzer.BackendQueryAnalyzer;
import ch.psi.daq.domain.query.channels.ChannelNameCache;
import ch.psi.daq.domain.query.channels.ChannelsRequest;
import ch.psi.daq.domain.query.channels.ChannelsResponse;
import ch.psi.daq.domain.query.processor.QueryProcessor;
import ch.psi.daq.queryrest.query.model.ChannelInfosStreamImpl;

public class QueryManagerImpl implements QueryManager {
   @SuppressWarnings("unused")
   private static final Logger LOGGER = LoggerFactory.getLogger(QueryManagerImpl.class);

   @Resource
   private ApplicationContext appContext;

   @Resource
   private Function<BackendQuery, BackendQueryAnalyzer> queryAnalizerFactory;

   @Resource(name = DomainConfig.BEAN_NAME_BACKEND_ACCESS)
   private BackendAccess backendAccess;

   @Resource(name = DomainConfig.BEAN_NAME_CHANNEL_NAME_CACHE)
   private ChannelNameCache channelNameCache;

   @PostConstruct
   public void afterPropertiesSet() {}

   @PreDestroy
   public void destroy() {}

   @Override
   public List<ChannelsResponse> getChannels(ChannelsRequest request) {
      // in case not specified use defaults (e.g. GET)
      if (request == null) {
         request = new ChannelsRequest();
      }

      return channelNameCache.getChannels(request);
   }

   public Collection<ChannelInfos> getChannelInfos(ChannelNameRequest request) {
      // set backends if not defined yet
      channelNameCache.configureBackends(request.getChannels());

      Stream<ChannelInfos> stream = request.getRequestsByBackend().entrySet().stream()
            .filter(entry ->
                  backendAccess.hasDataReader(entry.getKey())
                        && backendAccess.hasChannelInfoReader(entry.getKey()))
            .flatMap(entry -> {
               return entry.getValue().getChannelInfos(entry.getKey(), backendAccess)
                     .entrySet().stream()
                     .map(innerEntry -> {
                        return new ChannelInfosStreamImpl(
                              new ChannelName(innerEntry.getKey(), entry.getKey()),
                              innerEntry.getValue()
                           );
                        }
                     );
            });

      // materialize
      return stream.collect(Collectors.toList());
   }

   @Override
   public List<Entry<DAQQueryElement, Stream<Triple<BackendQuery, ChannelName, ?>>>> getEvents(DAQQueries queries) {
      // set backends if not defined yet
      channelNameCache.configureBackends(queries);

      List<Entry<DAQQueryElement, Stream<Triple<BackendQuery, ChannelName, ?>>>> results =
            new ArrayList<>(queries.getQueries().size());

      for (DAQQueryElement queryElement : queries) {
         Stream<Triple<BackendQuery, ChannelName, ?>> resultStreams =
               BackendQueryImpl
                     .getBackendQueries(queryElement)
                     .stream()
                     .filter(
                           query ->
                           backendAccess.hasDataReader(query.getBackend())
                                 && backendAccess.hasQueryProcessor(query.getBackend())
                     )
                     .flatMap(
                           query -> {
                              QueryProcessor processor = backendAccess.getQueryProcessor(query.getBackend());
                              BackendQueryAnalyzer queryAnalizer = queryAnalizerFactory.apply(query);

                              /* all the magic happens here */
                              Stream<Entry<ChannelName, Stream<? extends DataEvent>>> channelToDataEvents =
                                    processor.process(queryAnalizer);
                              /* do post-process */
                              Stream<Entry<ChannelName, ?>> channelToData =
                                    queryAnalizer.postProcess(channelToDataEvents);

                              return channelToData.map(entry -> {
                                 return Triple.of(query, entry.getKey(), entry.getValue());
                              });
                           });

         // Now we have a stream that loads elements sequential BackendQuery by BackendQuery.
         // By materializing the outer Stream the elements of all BackendQuery are loaded async
         // (speeds things up but requires also more memory - i.e. it relies on Backends not loading
         // all elements into memory at once)
         resultStreams = resultStreams.collect(Collectors.toList()).stream();

         results.add(Pair.of(queryElement, resultStreams));
      }

      return results;
   }

}
