package ch.psi.daq.test.queryrest.query;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Random;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.atomic.AtomicLong;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import java.util.stream.LongStream;
import java.util.stream.Stream;

import javax.annotation.Resource;

import org.apache.commons.lang3.ArrayUtils;

import com.google.common.collect.Lists;

import ch.psi.bsread.message.Type;
import ch.psi.daq.common.ordering.Ordering;
import ch.psi.daq.common.time.TimeUtils;
import ch.psi.daq.domain.DataEvent;
import ch.psi.daq.domain.FieldNames;
import ch.psi.daq.domain.backend.Backend;
import ch.psi.daq.domain.config.DomainConfig;
import ch.psi.daq.domain.events.ChannelConfiguration;
import ch.psi.daq.domain.events.ChannelEvent;
import ch.psi.daq.domain.events.MetaPulseId;
import ch.psi.daq.domain.events.impl.ChannelConfigurationImpl;
import ch.psi.daq.domain.events.impl.ChannelEventImpl;
import ch.psi.daq.domain.events.impl.MetaPulseIdImpl;
import ch.psi.daq.domain.json.channels.info.ChannelInfo;
import ch.psi.daq.domain.json.channels.info.ChannelInfoImpl;
import ch.psi.daq.domain.query.event.EventQuery;
import ch.psi.daq.domain.query.event.StreamEventQuery;
import ch.psi.daq.domain.query.range.PulseIdRangeQuery;
import ch.psi.daq.domain.query.range.TimeRangeQuery;
import ch.psi.daq.domain.reader.MetaStreamEventQuery;
import ch.psi.daq.domain.reader.StreamEventReader;
import ch.psi.daq.domain.test.backend.TestBackendAccess;
import ch.psi.daq.domain.test.gen.TestDataGen;
import ch.psi.daq.domain.utils.PropertiesUtils;
import ch.psi.data.stream.converters.impl.UShortByteValueConverter;

public abstract class AbstractStreamEventReader implements StreamEventReader<ChannelEvent> {
   private static final Random random = new Random(0);
   private static final int KEYSPACE = 1;
   private List<String> channels;
   private AtomicLong channelNameCallCounter = new AtomicLong();

   private TestDataGen dataGen;
   private Backend backend;
   private String testChannelName;

   @Resource(name = DomainConfig.BEAN_NAME_TEST_BACKEND_ACCESS)
   private TestBackendAccess testBackendAccess;


   public AbstractStreamEventReader() {
      this.channels = Lists.newArrayList(
            "BoolScalar",
            "BoolWaveform",
            "Int8Scalar",
            "Int8Waveform",
            "UInt8Scalar",
            "UInt8Waveform",
            "Int16Scalar",
            "Int16Waveform",
            "UInt16Scalar",
            "UInt16Waveform",
            "Int32Scalar",
            "Int32Waveform",
            "UInt32Scalar",
            "UInt32Waveform",
            "Int64Scalar",
            "Int64Waveform",
            "UInt64Scalar",
            "UInt64Waveform",
            "Float32Scalar",
            "Float32Waveform",
            "Float64Scalar",
            "Float64Waveform",
            "StringScalar");
   }

   protected void init(Backend backend) {
      this.backend = backend;
      this.dataGen = testBackendAccess.getTestDataGen(backend);
      this.testChannelName = backend.getKey() + "_TestChannel_";
   }

   @Override
   public Backend getBackend() {
      return backend;
   }

   /**
    * @{inheritDoc
    */
   @Override
   public Stream<String> getChannelStream(String regex) {
      channels.add(testChannelName + channelNameCallCounter.incrementAndGet());

      Stream<String> channelStream = channels.stream();
      if (regex != null) {
         Pattern pattern = Pattern.compile(regex.toLowerCase());
         channelStream = channelStream.filter(channel -> pattern.matcher(channel.toLowerCase()).find());
      }
      return channelStream;
   }

   // @Override
   // public Stream<ChannelEvent> getEventStream(PulseIdRangeQuery query) {
   // return getDummyEventStream(query.getChannel(), query.getStartPulseId(), query.getEndPulseId(),
   // query.getEventColumns())
   // .filter(query.getFilterOrDefault(EventQuery.NO_OP_FILTER));
   // }

   @Override
   public Stream<ChannelEvent> getEventStream(TimeRangeQuery query) {
      return getDummyEventStream(query.getChannel(), getBackend(), query.getStartMillis() / 10,
            query.getEndMillis() / 10)
            .filter(query.getFilterOrDefault(EventQuery.NO_OP_FILTER))
            // misuse value modification
            .peek(query.getValueTransformation());
   }

   public Stream<ChannelEvent> getEventStream(EventQuery eventQuery, Stream<? extends StreamEventQuery> queryProviders) {
      Stream<ChannelEvent> result = queryProviders.map(ceq -> {
         if (ceq instanceof MetaStreamEventQuery) {
            return getEvent((MetaStreamEventQuery<ChannelEvent>) ceq);
         } else {
            throw new UnsupportedOperationException("This is not yet implemented!");
         }
      });

      return result;
   }

   public static Stream<ChannelEvent> getDummyEventStream(String channelParam, Backend backend, long startIndex,
         long endIndex,
         String... columns) {
      String channelLower = channelParam.toLowerCase();
      String channel =
            (columns == null || columns.length == 0 || ArrayUtils.contains(columns, FieldNames.FIELD_CHANNEL)) ? channelParam
                  : null;

      LongStream millisRangeStream = null;

      if (channelParam.contains("[") && channelParam.contains("]")) {
         millisRangeStream =
               Arrays.stream(
                     channelParam.substring(
                           channelParam.indexOf("[") + 1,
                           channelParam.indexOf("]"))
                           .split(",")
                     )
                     .mapToLong(str -> Long.parseLong(str.trim()) * 10);
      } else if (channelParam.contains("{") && channelParam.contains("}")) {
         millisRangeStream =
               Arrays.stream(
                     channelParam.substring(
                           channelParam.indexOf("{") + 1,
                           channelParam.indexOf("}"))
                           .split(",")
                     )
                     .mapToLong(str -> Long.parseLong(str));
      } else {
         millisRangeStream = LongStream.rangeClosed(startIndex * 10, endIndex * 10)
               .filter(val -> val % 10 == 0);
      }

      Stream<ChannelEvent> eventStream =
            millisRangeStream.mapToObj(
                  millis -> {
                     long i = millis / 10;
                     BigDecimal iocTime =
                           (columns == null || columns.length == 0 || ArrayUtils.contains(columns,
                                 FieldNames.FIELD_IOC_TIME)) ? TimeUtils.getTimeFromMillis(millis, 0)
                                 : PropertiesUtils.DEFAULT_VALUE_DECIMAL;
                     BigDecimal globalTime =
                           (columns == null || columns.length == 0 || ArrayUtils.contains(columns,
                                 FieldNames.FIELD_GLOBAL_TIME)) ? TimeUtils.getTimeFromMillis(millis, 0)
                                 : PropertiesUtils.DEFAULT_VALUE_DECIMAL;
                     long pulseId =
                           (columns == null || columns.length == 0 || ArrayUtils.contains(columns,
                                 FieldNames.FIELD_PULSE_ID)) ? i : PropertiesUtils.DEFAULT_VALUE_BIGINT_PRIMITIVE;

                     if (channelLower.contains("waveform")) {
                        long[] value = random.longs(8).toArray();
                        value[0] = i;
                        value[1] = i;
                        return new ChannelEventImpl(
                              channel,
                              backend,
                              iocTime,
                              pulseId,
                              globalTime,
                              KEYSPACE,
                              value
                        );

                     } else if (channelLower.contains("image")) {
                        int x = 80;
                        int y = 40;
                        int[] shape = new int[] {x, y};
                        short[] value = new short[x * y];
                        IntStream.range(0, value.length)
                              .forEach(index -> value[index] = UShortByteValueConverter.convertVal(random.nextInt()));
                        value[0] = UShortByteValueConverter.convertVal((int) i);
                        value[1] = UShortByteValueConverter.convertVal((int) i);
                        return new ChannelEventImpl(
                              channel,
                              backend,
                              iocTime,
                              pulseId,
                              globalTime,
                              KEYSPACE,
                              value,
                              shape
                        );
                     } else {
                        return new ChannelEventImpl(
                              channel,
                              backend,
                              iocTime,
                              pulseId,
                              globalTime,
                              KEYSPACE,
                              i
                        );
                     }
                  });

      return eventStream;
   }

   private List<? extends DataEvent> getDummyEvents(String channel, long startIndex, long endIndex, String... columns) {
      return getDummyEventStream(channel, getBackend(), startIndex, endIndex, columns).collect(Collectors.toList());
   }

   /**
    * @{inheritDoc
    */
   @Override
   public List<String> getChannels() {
      return Lists.newArrayList(channels);
   }

   /**
    * @{inheritDoc
    */
   @Override
   public List<String> getChannels(String regex) {
      return Lists.newArrayList(channels).stream().filter(s -> {
         return s.contains(regex);
      }).collect(Collectors.toList());
   }

   /**
    * @{inheritDoc
    */
   @Override
   public ChannelEvent getEvent(MetaStreamEventQuery<ChannelEvent> queryInfo, String... columns) {
      if (queryInfo.getPulseId() > 0) {
         return (ChannelEvent) getDummyEvents(queryInfo.getChannel(), queryInfo.getPulseId(), queryInfo.getPulseId(),
               columns)
               .get(0);
      }
      return (ChannelEvent) getDummyEvents(queryInfo.getChannel(), queryInfo.getGlobalMillis() / 10,
            queryInfo.getGlobalMillis() / 10).get(0);
   }

   /**
    * @{inheritDoc
    */
   @Override
   public CompletableFuture<ChannelEvent> getEventAsync(MetaStreamEventQuery<ChannelEvent> queryInfo, String... columns) {
      // implement when needed
      throw new UnsupportedOperationException();
   }

   // /**
   // * @{inheritDoc
   // */
   // @Override
   // public Stream<? extends StreamEventQuery> getStreamEventQueryStream(PulseIdRangeQuery query) {
   //
   // return dataGen.generateMetaPulseId(
   // query.getStartPulseId(),
   // (query.getEndPulseId() - query.getStartPulseId() + 1),
   // i -> i * 10,
   // i -> 0,
   // i -> i,
   // query.getChannel())
   // .stream()
   // .map(metaPulse -> {
   // metaPulse.setKeyspace(KEYSPACE);
   // return metaPulse;
   // });
   // }

   public Stream<? extends StreamEventQuery> getStreamEventQueryStream(TimeRangeQuery query) {

      return dataGen.generateMetaTime(
            KEYSPACE,
            3600,
            query.getStartMillis() / 10,
            ((query.getEndMillis() - query.getStartMillis()) / 10 + 1),
            i -> i * 10,
            i -> 0,
            i -> i,
            getBackend(),
            query.getChannel()).stream();
   }

   // /**
   // * @{inheritDoc
   // */
   // @Override
   // public Stream<MetaPulseId> getMetaStream(PulseIdRangeQuery query) {
   //
   // return getStreamEventQueryStream(query).map(r -> {
   // return (MetaPulseId) r;
   // });
   //
   // }


   /**
    * @{inheritDoc
    */
   @Override
   public Stream<? extends MetaStreamEventQuery<ChannelEvent>> getMetaStream(TimeRangeQuery query) {

      return getStreamEventQueryStream(query).map(r -> {
         return (MetaStreamEventQuery<ChannelEvent>) r;
      });
   }

   @Override
   public Stream<ChannelEvent> getEventStream(Stream<? extends MetaStreamEventQuery<ChannelEvent>> queryInfos,
         String... columns) {
      return getEventStream(null, queryInfos);
   }

   @Override
   public Stream<ChannelConfiguration> getChannelConfiguration(TimeRangeQuery query) {
      List<ChannelConfiguration> configs = new ArrayList<>();

      BigDecimal time = query.getStartTime();
      configs.add(
            new ChannelConfigurationImpl(
                  query.getChannel(),
                  time,
                  TimeUtils.getMillis(time) / 10,
                  0,
                  Type.Int32.getKey(),
                  new int[] {1},
                  false,
                  ChannelConfiguration.DEFAULT_LOCAL_WRITE,
                  ChannelConfiguration.DEFAULT_BIN_SIZE_IN_MILLIS,
                  ChannelConfiguration.SPLIT_COUNT,
                  ChannelConfiguration.DEFAULT_SOURCE,
                  ChannelConfiguration.DEFAULT_MODULO,
                  ChannelConfiguration.DEFAULT_OFFSET,
                  Backend.SF_DATABUFFER));
      if (query.getEndMillis() > query.getStartMillis()) {
         time = query.getEndTime();
         configs.add(
               new ChannelConfigurationImpl(
                     query.getChannel(),
                     time,
                     TimeUtils.getMillis(time) / 10,
                     1,
                     Type.Int32.getKey(),
                     new int[] {1},
                     false,
                     ChannelConfiguration.DEFAULT_LOCAL_WRITE,
                     ChannelConfiguration.DEFAULT_BIN_SIZE_IN_MILLIS,
                     ChannelConfiguration.SPLIT_COUNT,
                     ChannelConfiguration.DEFAULT_SOURCE,
                     ChannelConfiguration.DEFAULT_MODULO,
                     ChannelConfiguration.DEFAULT_OFFSET,
                     Backend.SF_DATABUFFER));
      }

      if (Ordering.desc.equals(query.getOrdering())) {
         Collections.reverse(configs);
      }

      return configs.stream();
   }

   @Override
   public TimeRangeQuery getTimeRangeQuery(PulseIdRangeQuery query) {
      return new TimeRangeQuery(
            TimeUtils.getTimeFromMillis(query.getStartPulseId() * 10, 0),
            TimeUtils.getTimeFromMillis(query.getEndPulseId() * 10, 0),
            query);
   }

   @Override
   public Stream<? extends ChannelInfo> getChannelInfoStream(TimeRangeQuery query) {
      return getChannelConfiguration(query)
            .map(channelConfiguration -> new ChannelInfoImpl(channelConfiguration));
   }

   @Override
   public void truncateCache() {}

   @Override
   public CompletableFuture<MetaPulseId> getStartMetaPulseIdAsync(PulseIdRangeQuery query) {
      return CompletableFuture.completedFuture(new MetaPulseIdImpl(query.getChannel(), getBackend(), query
            .getStartPulseId(),
            TimeUtils.getTimeFromMillis(query.getStartPulseId() * 10, 0)));
   }

   @Override
   public CompletableFuture<MetaPulseId> getEndMetaPulseIdAsync(PulseIdRangeQuery query) {
      return CompletableFuture.completedFuture(new MetaPulseIdImpl(query.getChannel(), getBackend(), query
            .getEndPulseId(),
            TimeUtils.getTimeFromMillis(query.getEndPulseId() * 10, 0)));
   }
}
